@@ -7,6 +7,7 @@ from .healSparseCoverage import HealSparseCoverage
 from .utils import reduce_array, check_sentinel, _get_field_and_bitval, WIDE_NBIT, WIDE_MASK
 from .utils import is_integer_value, _compute_bitshift
 from .fits_shim import HealSparseFits, _make_header, _write_filename
+import warnings
 
 
 class HealSparseMap(object):
@@ -620,8 +621,8 @@ class HealSparseMap(object):
                     weight_values = None
 
                 if is_wide_mask:
-                    aux = pix_data.reshape((1, (nside_out//nside_coverage)**2, wmult, -1))
-                    aux = reduce_array(aux, reduction=reduction, axis=3).reshape((-1, wmult))
+                    aux = pix_data.reshape((1, (nside_out//nside_coverage)**2, -1, wmult))
+                    aux = reduce_array(aux, reduction=reduction, axis=2).reshape((-1, wmult))
                 elif is_rec_array:
                     aux = np.zeros(cov_map_out.nfine_per_cov, dtype=dtype_out)
                     for key, value in sparse_map_out.dtype.fields.items():
@@ -770,19 +771,57 @@ class HealSparseMap(object):
         # Fill with blank values
         self._sparse_map[oldsize:] = self._sparse_map[0]
 
+    def update_values_pos(self, ra_or_theta, dec_or_phi, values,
+                          lonlat=True, operation='replace'):
+        """
+        Update the values in the sparsemap for a list of positions.
+
+        Parameters
+        ----------
+        ra_or_theta : `float`, array-like
+           Angular coordinates of points on a sphere.
+        dec_or_phi : `float`, array-like
+           Angular coordinates of points on a sphere.
+        values : `np.ndarray`
+           Value or Array of values.  Must be same type as sparse_map.
+        lonlat : `bool`, optional
+           If True, input angles are longitude and latitude in degrees.
+           Otherwise, they are co-latitude and longitude in radians.
+        operation : `str`, optional
+           Operation to use to update values.  May be 'replace' (default),
+           'or', or 'and' (for bit masks)
+
+        Raises
+        ------
+        ValueError : Raised if positions do not resolve to unique
+           positions and operation is 'replace'.
+        """
+        return self.update_values_pix(hp.ang2pix(self._nside_sparse,
+                                                 ra_or_theta,
+                                                 dec_or_phi,
+                                                 lonlat=lonlat, nest=True),
+                                      values,
+                                      operation=operation)
+
     def update_values_pix(self, pixels, values, nest=True, operation='replace'):
         """
         Update the values in the sparsemap for a list of pixels.
+        The list of pixels must be unique if the operation is 'replace'.
 
         Parameters
         ----------
         pixels : `np.ndarray`
            Integer array of sparse_map pixel values
         values : `np.ndarray`
-           Value or Array of values.  Must be same type as sparse_map
+           Value or Array of values.  Must be same type as sparse_map.
         operation : `str`, optional
            Operation to use to update values.  May be 'replace' (default),
            'or', or 'and' (for bit masks)
+
+        Raises
+        ------
+        ValueError : Raised if positions do not resolve to unique
+           positions and operation is 'replace'.
         """
 
         if operation != 'replace':
@@ -791,6 +830,11 @@ class HealSparseMap(object):
             if not self.is_integer_map or self._sentinel != 0:
                 raise ValueError("Can only use and/or with integer map with 0 sentinel")
 
+        if operation == 'replace':
+            # Check for unique pixel positions
+            if len(np.unique(pixels)) < len(pixels):
+                raise ValueError("List of pixels must be unique if operation='replace'")
+
         # If _not_ recarray, we can use a single int/float
         is_single_value = False
         _values = values
@@ -857,16 +901,24 @@ class HealSparseMap(object):
             if operation == 'replace':
                 self._sparse_map[_pix[in_cov] + self._cov_map[ipnest_cov[in_cov]]] = _values[0]
             elif operation == 'or':
-                self._sparse_map[_pix[in_cov] + self._cov_map[ipnest_cov[in_cov]]] |= _values[0]
+                np.bitwise_or.at(self._sparse_map,
+                                 _pix[in_cov] + self._cov_map[ipnest_cov[in_cov]],
+                                 _values[0])
             elif operation == 'and':
-                self._sparse_map[_pix[in_cov] + self._cov_map[ipnest_cov[in_cov]]] &= _values[0]
+                np.bitwise_and.at(self._sparse_map,
+                                  _pix[in_cov] + self._cov_map[ipnest_cov[in_cov]],
+                                  _values[0])
         else:
             if operation == 'replace':
                 self._sparse_map[_pix[in_cov] + self._cov_map[ipnest_cov[in_cov]]] = _values[in_cov]
             elif operation == 'or':
-                self._sparse_map[_pix[in_cov] + self._cov_map[ipnest_cov[in_cov]]] |= _values[in_cov]
+                np.bitwise_or.at(self._sparse_map,
+                                 _pix[in_cov] + self._cov_map[ipnest_cov[in_cov]],
+                                 _values[in_cov])
             elif operation == 'and':
-                self._sparse_map[_pix[in_cov] + self._cov_map[ipnest_cov[in_cov]]] &= _values[in_cov]
+                np.bitwise_and.at(self._sparse_map,
+                                  _pix[in_cov] + self._cov_map[ipnest_cov[in_cov]],
+                                  _values[in_cov])
 
         # Update the coverage map for the rest of the pixels (if necessary)
         if out_cov.sum() > 0:
@@ -886,21 +938,25 @@ class HealSparseMap(object):
                     self._sparse_map[oldsize:][_pix[out_cov] + self._cov_map[ipnest_cov[out_cov]] -
                                                oldsize] = _values[0]
                 elif operation == 'or':
-                    self._sparse_map[oldsize:][_pix[out_cov] + self._cov_map[ipnest_cov[out_cov]] -
-                                               oldsize] |= _values[0]
+                    np.bitwise_or.at(self._sparse_map[oldsize:],
+                                     _pix[out_cov] + self._cov_map[ipnest_cov[out_cov]] - oldsize,
+                                     _values[0])
                 elif operation == 'and':
-                    self._sparse_map[oldsize:][_pix[out_cov] + self._cov_map[ipnest_cov[out_cov]] -
-                                               oldsize] &= _values[0]
+                    np.bitwise_and.at(self._sparse_map[oldsize:],
+                                      _pix[out_cov] + self._cov_map[ipnest_cov[out_cov]] - oldsize,
+                                      _values[0])
             else:
                 if operation == 'replace':
                     self._sparse_map[oldsize:][_pix[out_cov] + self._cov_map[ipnest_cov[out_cov]] -
                                                oldsize] = _values[out_cov]
                 elif operation == 'or':
-                    self._sparse_map[oldsize:][_pix[out_cov] + self._cov_map[ipnest_cov[out_cov]] -
-                                               oldsize] |= _values[out_cov]
+                    np.bitwise_or.at(self._sparse_map[oldsize:],
+                                     _pix[out_cov] + self._cov_map[ipnest_cov[out_cov]] - oldsize,
+                                     _values[out_cov])
                 elif operation == 'and':
-                    self._sparse_map[oldsize:][_pix[out_cov] + self._cov_map[ipnest_cov[out_cov]] -
-                                               oldsize] &= _values[out_cov]
+                    np.bitwise_and.at(self._sparse_map[oldsize:],
+                                      _pix[out_cov] + self._cov_map[ipnest_cov[out_cov]] - oldsize,
+                                      _values[out_cov])
 
     def set_bits_pix(self, pixels, bits, nest=True):
         """
@@ -948,11 +1004,12 @@ class HealSparseMap(object):
         value = self._sparse_map[0].copy()
         for bit in bits:
             field, bitval = _get_field_and_bitval(bit)
-            value[field] |= ~bitval
+            value[field] |= bitval
 
-        self.update_values_pix(pixels, value, nest=nest, operation='and')
+        # A bit reset is performed with &= ~(bit1 | bit2)
+        self.update_values_pix(pixels, ~value, nest=nest, operation='and')
 
-    def get_values_pos(self, theta_or_ra, phi_or_dec, lonlat=True, valid_mask=False):
+    def get_values_pos(self, ra_or_theta, dec_or_phi, lonlat=True, valid_mask=False):
         """
         Get the map value for the position.  Positions may be theta/phi
         co-latitude and longitude in radians, or longitude and latitude in
@@ -960,9 +1017,9 @@ class HealSparseMap(object):
 
         Parameters
         ----------
-        theta_or_ra : `float`, array-like
+        ra_or_theta : `float`, array-like
            Angular coordinates of points on a sphere.
-        phi_or_dec : `float`, array-like
+        dec_or_phi : `float`, array-like
            Angular coordinates of points on a sphere.
         lonlat : `bool`, optional
            If True, input angles are longitude and latitude in degrees.
@@ -975,7 +1032,7 @@ class HealSparseMap(object):
         values : `np.ndarray`
            Array of values/validity from the map.
         """
-        return self.get_values_pix(hp.ang2pix(self._nside_sparse, theta_or_ra, phi_or_dec,
+        return self.get_values_pix(hp.ang2pix(self._nside_sparse, ra_or_theta, dec_or_phi,
                                               lonlat=lonlat, nest=True),
                                    valid_mask=valid_mask)
 
@@ -1020,7 +1077,7 @@ class HealSparseMap(object):
             # Just return the values
             return values
 
-    def check_bits_pos(self, theta_or_ra, phi_or_dec, bits, lonlat=True):
+    def check_bits_pos(self, ra_or_theta, dec_or_phi, bits, lonlat=True):
         """
         Check the bits at the map for an array of positions.  Positions may be
         theta/phi co-latitude and longitude in radians, or longitude and
@@ -1028,9 +1085,9 @@ class HealSparseMap(object):
 
         Parameters
         ----------
-        theta_or_ra : `float`, array-like
+        ra_or_theta : `float`, array-like
            Angular coordinates of points on a sphere.
-        phi_or_dec : `float`, array-like
+        dec_or_phi : `float`, array-like
            Angular coordinates of points on a sphere.
         lonlat : `bool`, optional
            If True, input angles are longitude and latitude in degrees.
@@ -1045,7 +1102,7 @@ class HealSparseMap(object):
            set
         """
         return self.check_bits_pix(hp.ang2pix(self._nside_sparse,
-                                              theta_or_ra, phi_or_dec,
+                                              ra_or_theta, dec_or_phi,
                                               lonlat=lonlat, nest=True),
                                    bits)
 
@@ -1464,10 +1521,10 @@ class HealSparseMap(object):
         else:
             return hp.pix2ang(self.nside_sparse, self.valid_pixels, lonlat=lonlat, nest=True)
 
-    def degrade(self, nside_out, reduction='mean', weights=None):
+    def _degrade(self, nside_out, reduction='mean', weights=None):
         """
-        Reduce the resolution, i.e., increase the pixel size
-        of a given sparse map.
+        Auxiliary method to reduce the resolution, i.e., increase the pixel size
+        of a given sparse map (which is called by `degrade`).
 
         Parameters
         ----------
@@ -1527,8 +1584,8 @@ class HealSparseMap(object):
                 reduction operation, try and/or.')
             else:
                 nbits = self._sparse_map.shape[1]
-                aux = self._sparse_map.reshape((npop_pix+1, (nside_out//self.nside_coverage)**2, nbits, -1))
-                sparse_map_out = reduce_array(aux, reduction=reduction, axis=3).reshape((-1, nbits))
+                aux = self._sparse_map.reshape((npop_pix+1, (nside_out//self.nside_coverage)**2, -1, nbits))
+                sparse_map_out = reduce_array(aux, reduction=reduction, axis=2).reshape((-1, nbits))
                 sentinel_out = self._sentinel
 
         # Work with RecArray (we have to change the resolution to all maps...)
@@ -1580,6 +1637,47 @@ class HealSparseMap(object):
         return HealSparseMap(cov_map=new_cov_map, sparse_map=sparse_map_out,
                              nside_sparse=nside_out, primary=self._primary, sentinel=sentinel_out)
 
+    def degrade(self, nside_out, reduction='mean', weights=None):
+        """
+        Method to reduce the resolution, i.e., increase the pixel size
+        of a given sparse map.
+
+        Parameters
+        ----------
+        nside_out : `int`
+           Output Nside resolution parameter.
+        reduction : `str`
+           Reduction method (mean, median, std, max, min, and, or, sum, prod, wmean).
+        weights : `HealSparseMap`
+           If the reduction is `wmean` this is the map with the weights to use.
+           It should have the same characteristics as the original map.
+
+        Returns
+        -------
+        healSparseMap : `HealSparseMap`
+           New map, at the desired resolution.
+        """
+        if nside_out < self.nside_coverage:
+            # The way we do the reduction requires nside_out to be >= nside_coverage
+            # we allocate a new map with the required nside_out
+            # CAUTION: This may require a lot of memory!!
+            warnings.warn("`nside_out` < `nside_coverage`. \
+                            Allocating new map with nside_coverage=nside_out",
+                          ResourceWarning)
+            sparse_map_out = HealSparseMap.make_empty_like(self,
+                                                           nside_coverage=nside_out)
+            if weights is not None:
+                wgt_valid = weights.valid_pixels
+                _weights = HealSparseMap.make_empty_like(weights, nside_coverage=nside_out)
+                _weights[wgt_valid] = weights[wgt_valid]
+                weights = _weights
+            valid_pixels = self.valid_pixels
+            sparse_map_out[valid_pixels] = self[valid_pixels]
+            sparse_map_out = sparse_map_out._degrade(nside_out, reduction=reduction, weights=weights)
+        else:
+            sparse_map_out = self._degrade(nside_out, reduction=reduction, weights=weights)
+        return sparse_map_out
+
     def apply_mask(self, mask_map, mask_bits=None, mask_bit_arr=None, in_place=True):
         """
         Apply an integer mask to the map.  All pixels in the integer
@@ -1687,7 +1785,7 @@ class HealSparseMap(object):
         """
         Set part of a healpix map
         """
-        if isinstance(key, int):
+        if isinstance(key, numbers.Integral):
             # Set a single pixel
             return self.update_values_pix(np.array([key]), value)
         elif isinstance(key, slice):
